#' Find marker genes using smashpy
#'
#' @param sce A SingleCellExperiment object,
#' @param pars A list of parameters passed to the function.
#'
#' @return A list containing the time of the computation and results of the
#'  method.
#'
run_smash <- function(sce, pars) {

  # Ensure we use the correct conda environment.
  # FIXME: Move to .Rprofile.
  Sys.setenv(RETICULATE_PYTHON = "/home/jpullin/.conda/envs/mage/bin/python")
  library(reticulate)

  # IPython and cuda errors seem spurious.
  sm <- import("smashpy")
  sm <- sm$smashpy()

  # Needed for AnnData conversion.
  sc <- import("scanpy")

  # As in `run_scanpy` file.
  sce@metadata <- list()

  adata <- sc$AnnData(
    X   = t(as.matrix(logcounts(sce))),
    obs = as.data.frame(colData(sce)),
    var = as.data.frame(rowData(sce))
  )

  adata$obs["annotation"] <- colLabels(sce)

  # Steps taken (verbatim + R translation) from GitLab README.

  times <- system.time({

    # Data preparation
    sm$data_preparation(adata)

    # Removing general genes
    adata <- sm$remove_general_genes(adata)

    # Removing genes expressed in less than 30% within groups
    adata <- sm$remove_features_pct(adata, group_by = "annotation", pct = 0.3)

    # Removing genes expressed in more than 50% in a given group where genes are
    # expressed for more 75% within a given group
    adata <- sm$remove_features_pct_2groups(adata, group_by = "annotation",
                                            pct1 = 0.75, pct2 = 0.5)

    # Inverse PCA to remove unimportant genes
    adata <- sm$scale_filter_features(adata, n_components = NULL,
                                      filter_expression = TRUE)

    # Run deep neural network to locate optimal markers for classification of
    # cells according to the original user annotations
    sm$DNN(adata, group_by = "annotation", model = NULL, balance = TRUE,
           verbose = TRUE, save = FALSE)

    # Top 20 genes as a final dictionary, for each annotation (class) provided
    # Calculate the importances of each gene using the Shapley value.
    out <- sm$run_shap(adata, group_by = "annotation", model = NULL,
                       verbose = TRUE, pct = 0.1,
                       restrict_top = tuple("local", as.integer(20)))
  })

  raw_cluster_mgs <- out[[2]]

  cluster_mgs <- tibble::as_tibble(raw_cluster_mgs)

  result <- tidyr::pivot_longer(
    tibble::as_tibble(cluster_mgs),
    cols = everything(),
    names_to = "cluster",
    values_to = "gene"
  )

  result <- dplyr::arrange(result, cluster)

  result <- dplyr::mutate(
    result,
    log_fc = 0,
    raw_statistic = 0,
    scaled_statistic = 0,
    p_value = 0,
    p_value_adj = 0
  )

  # We use the 'elapsed' time.
  time <- times[[3]]

  validate_mgs_result(result)

  list(time = time, result = result, raw_result = raw_cluster_mgs, pars = pars)
}
